"""
This module provides the functionality for generating and rating attack path graphs
based on attack feasibility.
"""

from aranea.apg.compose_attack_path_graph import compose_attack_path_graph
from aranea.apg.get_shortest_path import get_shortest_path
from aranea.models.graph_model import (AttackPathGraph, ComponentNode,
                                       EcuClassificationName, Graph,
                                       GraphCollection)


def get_attack_paths(graph: Graph) -> GraphCollection[AttackPathGraph]:
    """
    Function for generating a collection of attack path graphs.
    :param graph: The graph to generate paths from.
    :type graph: Graph
    :return: The generated paths in a collection.
    :rtype: GraphCollection[AttackPathGraph]
    """
    entry_point_nodes = {
        key: value
        for key, value in graph.nodes.items()
        if isinstance(value, ComponentNode)
        and EcuClassificationName.ENTRY_POINT
        in [classification.name for classification in value.classifications]
    }

    target_nodes = {
        key: value
        for key, value in graph.nodes.items()
        if isinstance(value, ComponentNode)
        and EcuClassificationName.CRITICAL_ELEMENT
        in [classification.name for classification in value.classifications]
    }

    graph_collection: GraphCollection[AttackPathGraph] = GraphCollection(graphs=[])

    for entry_uuid, entry_component in entry_point_nodes.items():
        for target_uuid in target_nodes.keys():
            for technical_capability in entry_component.technical_capabilities:
                shortest_path = get_shortest_path(
                    graph, entry_uuid, target_uuid, technical_capability.name
                )

                if shortest_path:
                    graph_collection.graphs.append(
                        compose_attack_path_graph(
                            shortest_path[0],
                            shortest_path[1],
                            entry_uuid,
                            target_uuid,
                            technical_capability,
                        )
                    )

    return graph_collection


def get_path_afr_and_length(graph: AttackPathGraph, x: float) -> tuple[float, int]:
    """
    Function for rating an attack path graph and getting the length of the path.

    :param graph: The graph to rate
    :type graph: AttackPathGraph
    :param x: A user given input for factoring the path length impact.
    :type x: float
    :return: The rating and the length of the path.
    :rtype: tuple[float, int]
    """
    if len(graph.nodes) == 0 or len(graph.networks) == 0:
        raise ValueError("Empty graph")

    if x <= 0:
        raise ValueError("x must be positive")

    raw_feasibility: float = 1

    raw_feasibility *= graph.entry_technical_capability.feasibility_rating

    for node in graph.nodes.values():
        if isinstance(node, ComponentNode):
            for classification in node.classifications:
                raw_feasibility *= classification.feasibility_rating

    for network in graph.networks:
        raw_feasibility *= network.protocol_type.feasibility_rating

    count_hops = len(graph.nodes) - 1
    omega_hops = count_hops * x
    afr_path = raw_feasibility / omega_hops

    return afr_path, count_hops
