"""
This module provides functions for composing attack path graphs.
"""

from typing import Annotated, cast
from uuid import UUID

from pydantic.types import UuidVersion

from aranea.models.graph_model import (AttackPathGraph, ComponentNode, Network,
                                       NodeUnionType, TechnicalCapability,
                                       TextOrientation)


def compose_attack_path_graph(
    nodes: dict[Annotated[UUID, UuidVersion(4)], ComponentNode],
    networks: list[Network],
    start: UUID,
    end: UUID,
    entry_technical_capability: TechnicalCapability,
) -> AttackPathGraph:
    """
    Function that composes an attack path graph from nodes and networks.

    :param nodes: The nodes to use.
    :type nodes: dict[Annotated[UUID, UuidVersion(4)], ComponentNode]
    :param networks: The networks to use.
    :type networks: list[Network]
    :param start: The start node to use.
    :type start: UUID
    :param end: The end node to use.
    :type end: UUID
    :param entry_technical_capability: The technical capability to use.
    :type entry_technical_capability: TechnicalCapability
    :return: The composed attack path graph.
    :rtype: AttackPathGraph
    """
    return AttackPathGraph(
        label=(f"AttackPathGraph-{start}-{end}", TextOrientation.HORIZONTAL, 1),
        nodes=cast(dict[UUID, NodeUnionType], nodes),
        networks=networks,
        entry_technical_capability=entry_technical_capability,
        entry_ecu=start,
        target_ecu=end,
    )
