"""
Module to extract network line labels from the PDF document.
"""

import logging
from typing import Annotated, Any, cast
from uuid import UUID

from pymupdf import Page
from shapely import LineString, Point, Polygon
from tqdm import tqdm
from tqdm.contrib.logging import logging_redirect_tqdm

from aranea.models.graph_model import (ComponentNode, Edge, Network,
                                       NodeUnionType, Text, TextOrientation,
                                       WaypointNode, XorNode, get_default_text)
from aranea.p2g.util import REM, DictTextBlock, gendocstring

logger = logging.getLogger(__file__)

HORIZONTAL: tuple[float, float] = (1.0, 0.0)
TEXT_TO_IGNORE = ["xor", "...", ". . ."]


def __is_horizontal(x0: float, y0: float, x1: float, y1: float) -> bool:
    return abs(x1 - x0) > abs(y1 - y0)


def __node_attachment2point(
    rem: REM, node: NodeUnionType, attach_x: float, attach_y: float
) -> Point:
    if isinstance(node, (ComponentNode, XorNode)):
        return Point(
            node.xRemFactor * rem + node.widthRemFactor * rem * attach_x,
            node.yRemFactor * rem + node.heightRemFactor * rem * attach_y,
        )
    if isinstance(node, WaypointNode):
        return Point(node.xRemFactor * rem, node.yRemFactor * rem)

    # we cannot correctly calculate the absolute point for a TextNode
    # attachment because TextNodes do miss height and width attributes.
    raise NotImplementedError(
        f"Cannot convert attachment point to absolute point for {node.__class__}"
    )


def __bbox2polygon(bbox: tuple[float, float, float, float]) -> Polygon:
    x1, y1, x2, y2 = bbox
    return Polygon(((x1, y1), (x2, y1), (x2, y2), (x1, y2), (x1, y1)))


def __is_text_line_to_be_ignored(line: dict[str, Any], outer_ecu_labels: list[str]) -> bool:
    for span in line["spans"]:
        if span["text"].lower().strip() in TEXT_TO_IGNORE:
            logger.debug('Skipping "%s" text span', span["text"])
            return True

        # TODO: There is still an issue with "(AMG only)" text
        # spans. We cannot just skip these text spans bc this string
        # can also occur in network labels

        if any(ol == span["text"].replace("*", "").strip() for ol in outer_ecu_labels):
            logger.debug('Skipping outer ECU label "%s"', span["text"])
            return True
    return False


# order edge candidates by orientation and distance to text span
def __sorting_key(
    line: LineString, span_poly: Polygon, text_orientation: TextOrientation
) -> tuple[bool, float]:
    p1, p2 = list(line.coords)
    line_is_horizontal = __is_horizontal(*p1, *p2)
    same_orientation = line_is_horizontal and text_orientation == TextOrientation.HORIZONTAL

    # we prefer edges with same orientation so invert `same_orientation` bc False < True
    return (not same_orientation, line.distance(span_poly))


# TODO: maybe prefer edge candidates that do not already have a text span associated with them
def __find_nearest_edge_to_text_line(
    text_line: dict[str, Any],
    max_label_to_line_distance: float,
    networks: list[Network],
    nodes: dict[UUID, NodeUnionType],
    waypoints: dict[UUID, WaypointNode],
    text_orientation: TextOrientation,
    rem: REM,
) -> Edge | None:
    text_line_poly = __bbox2polygon(text_line["bbox"])
    text_line_poly_for_intersection = text_line_poly.buffer(max_label_to_line_distance)
    edge_candidates: list[tuple[Edge, LineString]] = []
    for network in networks:
        for edge in network.edges:
            source = nodes.get(edge.sourceId, waypoints.get(edge.sourceId))
            target = nodes.get(edge.targetId, waypoints.get(edge.targetId))

            if source is None:
                raise ValueError("Edge source not found")
            if target is None:
                raise ValueError("Edge target not found")

            p1 = __node_attachment2point(
                rem,
                source,
                edge.sourceAttachmentPointX or 0,
                edge.sourceAttachmentPointY or 0,
            )
            p2 = __node_attachment2point(
                rem,
                target,
                edge.targetAttachmentPointX or 0,
                edge.targetAttachmentPointY or 0,
            )
            edge_line = LineString((p1, p2))

            if edge_line.intersects(text_line_poly_for_intersection):
                if all(
                    node.polygon(rem=rem).intersection(text_line_poly).is_empty
                    for node in nodes.values()
                    if isinstance(node, ComponentNode)
                ):
                    edge_candidates.append((edge, edge_line))

    if len(edge_candidates) > 0:
        return min(
            edge_candidates, key=lambda c: __sorting_key(c[1], text_line_poly, text_orientation)
        )[0]
    return None


def __edge_text_lines2text(
    text_lines: list[tuple[dict[str, Any], TextOrientation]], rem: REM
) -> Text:
    if text_lines[0][1] == TextOrientation.HORIZONTAL:
        text_lines = sorted(text_lines, key=lambda line: line[0]["bbox"][:1:-1])
    elif text_lines[0][1] == TextOrientation.VERTICAL:
        text_lines = sorted(text_lines, key=lambda line: line[0]["bbox"][:1])

    edge_label_text = "\n".join(
        ["".join([s["text"] for s in line[0]["spans"]]) for line in text_lines]
    )

    if len({line[1] for line in text_lines}) > 1:
        logger.warning('Multiple text orientations in edge label "%s"', edge_label_text)

    o = text_lines[0][1]
    s = text_lines[0][0]["spans"][0]["size"]
    return get_default_text(edge_label_text, text_orientation=o, rem_factor=s / rem)


@gendocstring
def add_labels_to_networks(
    rem: REM,
    page: Annotated[Page, "The page containing the desired labels"],
    networks: Annotated[list[Network], "The list of networks"],
    nodes: Annotated[dict[UUID, NodeUnionType], "A dictionary of nodes"],
    waypoints: Annotated[dict[UUID, WaypointNode], "A dictionary of waypoints"],
    max_label_to_line_distance: Annotated[
        float, "The maximum distance between a network label and network line segment"
    ],
) -> Annotated[list[Network], "The list of networks with labels"]:
    """
    Finds network labels and assigns them to network edges.
    """

    word_blocks: list[DictTextBlock] = cast(
        list[DictTextBlock], page.get_text("dict")["blocks"]  # pyright: ignore
    )

    outer_ecu_labels: list[str] = [
        n.outerText[0]
        for n in nodes.values()
        if isinstance(n, ComponentNode) and n.outerText is not None
    ]

    # keep track of all spans to assign to an edge
    edge2text_lines: dict[Edge, list[tuple[dict[str, Any], TextOrientation]]] = {}
    with logging_redirect_tqdm():

        for block in tqdm(word_blocks, colour="BLUE", desc="Assigning network labels", leave=False):
            if "lines" not in block:
                continue
            for line in block["lines"]:
                if "spans" not in line:
                    continue

                text_orientation: TextOrientation = (
                    TextOrientation.HORIZONTAL
                    if line.get("dir") == HORIZONTAL
                    else TextOrientation.VERTICAL
                )
                if __is_text_line_to_be_ignored(line, outer_ecu_labels):
                    continue

                nearest_edge = __find_nearest_edge_to_text_line(
                    line,
                    max_label_to_line_distance,
                    networks,
                    nodes,
                    waypoints,
                    text_orientation,
                    rem,
                )
                if nearest_edge is None:
                    continue
                if nearest_edge in edge2text_lines:
                    edge2text_lines[nearest_edge].append((line, text_orientation))
                else:
                    edge2text_lines[nearest_edge] = [(line, text_orientation)]

    for edge, lines in edge2text_lines.items():
        if len(lines) > 0:
            edge.text = __edge_text_lines2text(lines, rem)

    return networks
