"""CLI module for generating attack paths for a provided architecture."""

import logging
import os
import sys
from pathlib import Path

import click

from aranea.apg import get_attack_paths
from aranea.models.graph_model import (AttackPathGraph, Graph, GraphCollection,
                                       GraphType)
from cli.lib.cli_context import EPILOG, CliContext
from cli.lib.utils import (export_drawio_file, graph_collection_to_drawio_file,
                           graph_collection_to_json_file)

logger: logging.Logger = logging.getLogger(__name__)


@click.command(epilog=EPILOG)
@click.argument(
    "architecture_json_path",
    nargs=1,
    required=True,
    type=click.Path(
        exists=True,
        file_okay=True,
        dir_okay=False,
        readable=True,
        resolve_path=True,
        path_type=Path,
    ),
)
@click.option(
    "-g",
    "--graph-number",
    "graph_number",
    default=0,
    show_default=True,
    type=int,
    help="Defines the graph number from the GraphCollection to generate the attack paths from.\
 Start counting from 0.",
)
@click.option(
    "--drawio-file/--no-drawio-file",
    "drawio_file",
    is_flag=True,
    show_default=True,
    default=True,
    required=False,
    type=bool,
    help="If set, every attack path graph will be exported as a page to a DrawIO file.",
)
@click.option(
    "-f",
    "--export-format",
    "export_format",
    type=click.Choice(["pdf", "png", "jpg", "svg", "vsdx", "xml"], case_sensitive=True),
    help="If set, the detected architecture is exported to the specified format. The export will\
 be disabled if the drawio_path is not set or the '--no-drawio-file' flag is set.\
 If you are using ``--export-format`` under Linux and EUID is 0 (root), drawio will be\
 run with ``--no-sandbox``. If the environment variable ``CI`` is also set to ``true`` it will\
 be run with ``xvfb-run``. This requires that ``xvfb`` is installed on the system.",
)
@click.option(
    "-n",
    "--output-file-name",
    "output_file_name",
    default="aranea_result",
    show_default=True,
    type=str,
    help="You can specify a custom name for the output file.",
)
@click.pass_obj
def apg(
    ctx: CliContext,
    architecture_json_path: Path,
    graph_number: int,
    drawio_file: bool,
    export_format: str | None,
    output_file_name: str,
) -> None:
    """Creates an ``GraphCollection`` JSON file with several ``AttackPathGraph``s from an existing
    ``GraphCollection`` JSON file. You need to specify which ``Graph`` from the ``GraphCollection``
    you want to generate the attack paths from. Default is the first item (0) of the ``graphs``
    list of the ``GraphCollection``.

    The provided JSON file must be a valid Aranea ``GraphCollection``.
    """

    logger.debug("---Starting the apg command---")

    with open(architecture_json_path, "r", encoding="UTF-8") as f:
        graph_collection: GraphCollection[Graph | AttackPathGraph] = (
            GraphCollection.model_validate_json(f.read())
        )

    check_valid_graph_index(graph_collection, graph_number)

    result: GraphCollection[AttackPathGraph] = get_attack_paths(
        graph_collection.graphs[graph_number]
    )

    json_file_path: str = os.path.join(ctx.user_config.output_dir, output_file_name + ".json")
    graph_collection_to_json_file(result, json_file_path)  # pyright: ignore

    if drawio_file:
        drawio_file_path: str = os.path.join(
            ctx.user_config.output_dir, output_file_name + ".drawio"
        )
        graph_collection_to_drawio_file(
            result, ctx.user_config.g2d, drawio_file_path  # pyright: ignore
        )

        if export_format is not None:
            export_drawio_file(
                drawio_path=ctx.user_config.drawio_path,
                drawio_file_path=drawio_file_path,
                export_format=export_format,
                output_dir=ctx.user_config.output_dir,
                output_file_name=output_file_name,
            )

    logger.debug("---Finished the ge command---")


def check_valid_graph_index(
    graph_collection: GraphCollection[GraphType], graph_number: int
) -> None:
    """Checks if the provided graph index is valid.

    If the provided graph index is not in the ``graphs`` list of the ``GraphCollection``, the
    program will exit with an error message.

    :param graph_collection: The ``GraphCollection`` to check the graph index against.
    :type graph_collection: GraphCollection
    :param graph_number: The graph index to check.
    :type graph_number: int
    """

    if graph_number >= len(graph_collection.graphs):
        logger.error(
            "The provided graph index %s is not in the ``graphs`` list of the ``GraphCollection``. "
            "Please provide a valid number.",
            graph_number,
        )
        sys.exit(1)
    if not isinstance(graph_collection.graphs[graph_number], Graph):  # pyright: ignore
        logger.error(
            "The provided graph index is not a ``Graph``. Please provide a valid ``Graph``."
        )
        sys.exit(1)
